"""
data_mov_att.py: other data movement instructions for AT&T
"""

from assembler.errors import check_num_args, InvalidConVal, InvalidArgument
from assembler.tokens import Instruction, IntegerTok, Register, Address


def check_source_val(instr, ops, data_type, line_num):
    """
    Checks if the constant value matches the transfer size

    Args:
        instr: Instruction name
        ops: Operand list
        data_type: Transfer size
                   'b': byte
                   'w': word
                   'l': long

    Raises an Invalid Constant Value error if mismatch found
    """
    if (not isinstance(ops[1], IntegerTok) and not isinstance(ops[1], Register)
            and not isinstance(ops[1], Address)):
        raise InvalidArgument("Invalid data", line_num)
    val = ops[1].get_val(line_num)
    if data_type == "b":
        if (val >= 2 ** 8 or val <= -(2 ** 8)):
            raise InvalidConVal(str(val), line_num)
    elif data_type == "w":
        if (val >= 2 ** 16 or
                ops[1].get_val(line_num) <= -(2 ** 16)):
            raise InvalidConVal(str(ops[1].get_val(line_num)), line_num)
    else:
        if (ops[1].get_val(line_num) >= 2 ** 32 or
                ops[1].get_val(line_num) <= -(2 ** 32)):
            raise InvalidConVal(str(ops[1].get_val(line_num)), line_num)


class Movb(Instruction):
    """
        <instr>
             movb
        </instr>
        <syntax>
            MOVB con, mem
        </syntax>
        <descr>
            Copies the value of op1 to the location mentioned in op2.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        check_source_val(self.get_nm(), ops, 'b', line_num)
        ops[0].set_val(ops[1].get_val(line_num), line_num)
        if isinstance(ops[0], Register):
            vm.changes.add(ops[0].get_nm())
        elif isinstance(ops[0], Address):
            vm.changes.add(f'MEM{ops[0].get_mem_addr(line_num)}')


class Movw(Instruction):
    """
        <instr>
             movw
        </instr>
        <syntax>
            MOVW con, mem
        </syntax>
        <descr>
            Copies the value of op1 to the location mentioned in op2.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        check_source_val(self.get_nm(), ops, 'w', line_num)
        ops[0].set_val(ops[1].get_val(line_num), line_num)
        if isinstance(ops[0], Register):
            vm.changes.add(ops[0].get_nm())
        elif isinstance(ops[0], Address):
            vm.changes.add(f'MEM{ops[0].get_mem_addr(line_num)}')


class Movl(Instruction):
    """
        <instr>
             movl
        </instr>
        <syntax>
            MOVL con, mem
        </syntax>
        <descr>
            Copies the value of op1 to the location mentioned in op2.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        check_source_val(self.get_nm(), ops, 'l', line_num)
        ops[0].set_val(ops[1].get_val(line_num), line_num)
        if isinstance(ops[0], Register):
            vm.changes.add(ops[0].get_nm())
        elif isinstance(ops[0], Address):
            vm.changes.add(f'MEM{ops[0].get_mem_addr(line_num)}')



'''
Arithmetic and Logical instructions for
RISC-V
'''
import operator as opfunc

from assembler.errors import check_num_args, IncorrectImmLength
from assembler.tokens import Instruction, MAX_INT
# from assembler.ops_check import one_op_arith
from .argument_check import check_reg_only, check_immediate_three
from .argument_check import check_immediate_two


def three_op_arith_reg(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    ops[0].set_val(check_overflow(operator(ops[1].get_val(line_num),
                   ops[2].get_val(line_num)), vm), line_num)
    vm.changes.add(ops[0].get_nm())


def three_op_arith_immediate(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    ops[0].set_val(check_overflow(operator(ops[1].get_val(line_num),
                   ops[2].get_val(line_num)), vm), line_num)
    vm.changes.add(ops[0].get_nm())


def get_three_ops(instr, ops, line_num):
    '''
    Function to grab the values in registers.
    Put in a separate so that we don't have to write out the checks
    all the time.

    '''
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    return (ops[0], ops[1], ops[2])


def get_three_ops_imm(instr, ops, line_num):
    '''
    Same concept as the function above.
    '''
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    return (ops[0], ops[1], ops[2])


def get_two_op_imm(instr, ops, line_num):
    """
    Again, same as above, except for reg, imm format
    """
    check_num_args(instr, ops, 2, line_num)
    check_immediate_two(instr, ops, line_num)
    return (ops[0], ops[1])


def check_overflow(val, vm):
    '''
    To emulate the wraparound that occurs when a number
    has too many bits to represent in machine code.

    '''
    if (val > MAX_INT):
        val = val - MAX_INT + 1
    return val


class Add(Instruction):
    """
        <instr>
             ADD
        </instr>
        <syntax>
            ADD reg, reg, reg
        </syntax>
        <descr>
            Performs an add operation on the two source registers and stores
            the result in the desination register. Overflows are ignored.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.add, line_num)


class Addi(Instruction):
    """
        <instr>
             ADDI
        </instr>
        <syntax>
            ADDI reg, reg, con
        </syntax>
        <descr>
            Adds the sign extended 12-bit immediate to the source register
            and stores the result in the destination register. Overflows are
            ignored.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.add, line_num)


class Sub(Instruction):
    """
        <instr>
            SUB
        </instr>
        <syntax>
            SUB reg, reg, reg
        </syntax>
        <descr>
            Performs a subtract operation on the two source registers and
            stores the result in the destination register. Overflows are
            ignored.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.sub, line_num)


class Mul(Instruction):
    """
        <instr>
            MUL
        </instr>
        <syntax>
            MUL reg, reg, reg
        </syntax>
        <descr>
            Performs a multiply operation on the two source registers and
            stores the lower 32 bits of the result in the destination register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.mul, line_num)


class And(Instruction):
    """
        <instr>
            AND
        </instr>
        <syntax>
            AND reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise AND operation on the two source registers and
            stores the result in the destination register. If the bit in both
            source registers is 1, then the destination register gets a 1;
            otherwise it gets a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.and_, line_num)


class Andi(Instruction):
    """
        <instr>
            ANDI
        </instr>
        <syntax>
            ANDI reg, reg, con
        </syntax>
        <descr>
            Performs a bitwise AND operation on a source registers and a 12-bit
            sign extended immediate and stores the result in the destination
            register. If the bit in both the source register and the immediate
            is 1, then the destination register gets a 1; otherwise it gets a
            0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.and_, line_num)


class Or(Instruction):
    """
        <instr>
            OR
        </instr>
        <syntax>
            OR reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise OR operation on two source registers and stores
            the result in the destination register. If the bit in both source
            registers is a 0, then the destination register gets a 0; otherwise
            it gets a 1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.or_, line_num)


class Ori(Instruction):
    """
        <instr>
            ORI
        </instr>
        <syntax>
            ORI reg, reg, con
        </syntax>
        <descr>
            Performs a bitwise OR operation on a source register and a 12-bit
            sign extended immediate and stores the result in the destination
            register. If the bit in both the source register and the immediate
            is 0, then the destination register gets a 0; otherwise it gets a
            1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.or_, line_num)


class Xor(Instruction):
    """
        <instr>
            XOR
        </instr>
        <syntax>
            XOR reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise XOR operation on two source registers and stores
            the result in the destination register. If the bit in the source
            registers is different, the destination register gets a 1;
            otherwise it gets a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.xor, line_num)


class Xori(Instruction):
    """
        <instr>
            XORI
        </instr>
        <syntax>
            XORI reg, reg, con
        </syntax>
        <descr>
            Performs a bitwise XOR operation on a source register and a 12-bit
            sign extended immediate and stores the result in the destination
            register. If the bit in the source register and the immediate is
            different, the destination register gets a 1; otherwise it gets a
            0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.xor, line_num)


class Srl(Instruction):
    """
        <instr>
            SRL
        </instr>
        <syntax>
            SRL reg, reg, reg
        </syntax>
        <descr>
            Performs a logical right shift on the register rs1 by the amount
            specified by the lower 5 bits of rs2. The result is stored in
            register rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        fixed_op2 = ops[2].get_val(line_num) % 32
        ops[0].set_val(check_overflow(opfunc.rshift(ops[1].get_val(line_num),
                       fixed_op2), vm), line_num)
        vm.changes.add(ops[0].get_nm())


class Srli(Instruction):
    """
        <instr>
            SRLI
        </instr>
        <syntax>
            SRLI reg, reg, con
        </syntax>
        <descr>
            Performs a logical right shift on the register rs1 by the amount
            specified by the 12 bit immediate. The result is stored in register
            rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.rshift, line_num)


class Sll(Instruction):
    """
        <instr>
            SLL
        </instr>
        <syntax>
            SLL reg, reg, reg
        </syntax>
        <descr>
            Performs a logical left shift on the register rs1 by the amount
            specified by the lower 5 bits of register rs2. The result is stored
            in register rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        fixed_op2 = ops[2].get_val(line_num) % 32
        ops[0].set_val(check_overflow(opfunc.lshift(ops[1].get_val(line_num),
                       fixed_op2), vm), line_num)
        vm.changes.add(ops[0].get_nm())


class Slli(Instruction):
    """
        <instr>
            SLLI
        </instr>
        <syntax>
            SLLI reg, reg, con
        </syntax>
        <descr>
            Perfors a logical left shift on the register rs1 by the amount
            specified by the 12 bit immediate. The result is stored in register
            rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.lshift, line_num)


class Slt(Instruction):
    """
        <instr>
            SLT
        </instr>
        <syntax>
            SLT reg, reg, reg
        </syntax>
        <descr>
                Compares the values in rs1 and rs2 and stores a 1 in rd if rs1
                is less than rs2. Otherwise rd gets a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        if (op2.get_val(line_num) - op3.get_val(line_num)) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Sltu(Instruction):
    """
        <instr>
            SLTU
        </instr>
        <syntax>
            SLTU reg, reg, reg
        </syntax>
        <descr>
            Performs an unsigned compare of the values in rs1 and rs2 and
            stores a 1 in rd if rs1 is less than rs2. Otherwise rd gets a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        abs_op2 = abs(op2.get_val(line_num))
        abs_op3 = abs(op3.get_val(line_num))
        if (abs_op2 - abs_op3) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Slti(Instruction):
    """
        <instr>
            SLTI
        </instr>
        <syntax>
            SLTI reg, reg, con
        </syntax>
        <descr>
            Compares the value in rs1 and the value of the 12-bit sign extended
            immediate. If rs1 is less than the immediate, rd gets a 1,
            otherwise it gets a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        if (op2.get_val(line_num) - op3.get_val(line_num)) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Sltiu(Instruction):
    """
        <instr>
            SLTIU
        </instr>
        <syntax>
            SLTIU reg, reg, con
        </syntax>
        <descr>
            Performs an unsigned compare of the value in rs1 and the 12-bit
            sign extended immediate. If rs1 is less than the immediate, rd gets
            a 1, otherwise it gets a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        abs_op2 = abs(op2.get_val(line_num))
        abs_op3 = abs(op3.get_val(line_num))
        if (abs_op2 - abs_op3) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Sra(Instruction):
    """
        <instr>
            SRA
        </instr>
        <syntax>
            SRA reg, reg, reg
        </syntax>
        <descr>
            Performs an arithmetic right shift on the register rs1 by the
            amount specified by the lower 5 bits of rs2. The result is stored
            in register rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        bin_str_op2 = bin(abs(op2.get_val(line_num)))[2:]
        if op2.get_val(line_num) >= 0:
            sign = '0'
        else:
            sign = '1'
        signed_str = sign * op3.get_val(line_num)
        shifted_str = signed_str + bin_str_op2[: - op3.get_val(line_num)]
        op1.set_val(int(shifted_str, 2), line_num)
        vm.changes.add(op1.get_nm())


class Srai(Instruction):
    """
        <instr>
            SRAI
        </instr>
        <syntax>
            SRAI reg, reg, imm
        </syntax>
        <descr>
            Performs an arithmetic right shift on the register rs1 by the
            amount specified by the 12-bit immediate. The result is stored in
            register rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        bin_str_op2 = bin(abs(op2.get_val(line_num)))[2:]
        if op2.get_val(line_num) >= 0:
            sign = '0'
        else:
            sign = '1'
        signed_str = sign * op3.get_val(line_num)
        shifted_str = signed_str + bin_str_op2[: - op3.get_val(line_num)]
        op1.set_val(int(shifted_str, 2), line_num)
        vm.changes.add(op1.get_nm())


class Div(Instruction):
    """
    <instr>
        DIV
    </instr>
    <syntax>
        DIV reg, reg, reg
    </syntax>
    <descr>
        Performs a divide using register rs1 as the dividend and register rs2
        as the divisor. The result is stored in register rd. To get the
        remainder, use REM.
    </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.floordiv, line_num)


class Divu(Instruction):
    """
    <instr>
        DIVU
    </instr>
    <syntax>
        DIVU reg, reg, reg
    </syntax>
    <descr>
        Performs an unsigned divide using register rs1 as the dividend and
        register rs2 as the divisor. The result is stored in register rd. To
        get the remainder, use REMU.
    </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        ops[0].set_val(
            check_overflow(
                opfunc.floordiv(
                    abs(ops[1].get_val(line_num)),
                    abs(ops[2].get_val(line_num))
                ), vm
            ),
            line_num
        )
        vm.changes.add(ops[0].get_nm())


class Rem(Instruction):
    """
    <instr>
        REM
    </instr>
    <syntax>
        REM reg, reg, reg
    </syntax>
    <descr>
        Performs a divide using register rs1 as the dividend and register rs2
        as the divisor. The remainder of the divide is stored in register rd.
        To get the result, use DIV.
    </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.mod, line_num)


class Remu(Instruction):
    """
    <instr>
        REMU
    </instr>
    <syntax>
        REMU reg, reg, reg
    </syntax>
    <descr>
        Performs an unsigned divide using register rs1 as the dividend and
        register rs2 as the divisor. The remainder of the divide is stored in
        register rd. To get the result, use DIVU.
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        ops[0].set_val(
            check_overflow(
                opfunc.mod(
                       abs(ops[1].get_val(line_num)),
                       abs(ops[2].get_val(line_num))
                ), vm),
            line_num
        )
        vm.changes.add(ops[0].get_nm())


class Lui(Instruction):
    """
    <instr>
        LUI
    </instr>
    <syntax>
        LUI reg, imm
    </syntax>
    <desc>
        Places the specified 20-bit immediate in the upper 20 bits of the
        specified register. Fills the lower 12-bits with zeroes.
    </desc>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2) = get_two_op_imm(self.get_nm(), ops, line_num)
        if op2.get_val(line_num) > 1048576:
            raise IncorrectImmLength(op2.get_val(line_num), line_num)
        op1.set_val(check_overflow(opfunc.lshift(op2.get_val(line_num),
                    12), vm), line_num)
        vm.changes.add(op1.get_nm())



"""
interrupts.py: interrupts in program
"""

from assembler.errors import ExitProg, check_num_args
from assembler.tokens import Instruction


class Syscall(Instruction):
    """
        <instr>
             SYSCALL
        </instr>
        <syntax>
            SYSCALL
        </syntax>
        <descr>
            Exits program
        </descr>
    """

    def fhook(self, ops, vm, line_num):
        check_num_args("SYSCALL", ops, 0, line_num)
        raise ExitProg(self.get_nm())



from assembler.errors import check_num_args


def checkFloat(ops, line_num):
    return isinstance(ops[0].get_val(line_num),
                      float) and isinstance(ops[1].get_val(line_num), float)


def get_one_op(instr, ops, line_num):
    check_num_args(instr, ops, 1, line_num)
    return ops[0]


def get_two_ops(instr, ops, line_num):
    check_num_args(instr, ops, 2, line_num)
    return (ops[0], ops[1])


def one_op_arith(ops, vm, instr, line_num, operator):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 1, line_num)
    ops[0].set_val(operator(ops[0].get_val(line_num)), line_num)
    vm.changes.add(ops[0].get_nm())



from ipykernel.kernelbase import Kernel
from assembler.virtual_machine import MIPSMachine
from assembler.assemble import assemble


class Mips_asmKernel(Kernel):
    implementation = 'mips_asm_kernel'
    implementation_version = '1.0'
    language = 'mips_asm'
    language_version = '1.0'
    language_info = {
        'name': 'mips_asm',
        'mimetype': 'mips_asm',
        'file_extension': 'x86',
    }
    banner = "Mips_asm kernel - run mips_asm assembly language"
    vm_machine = None

    def has_one_data_first(self, code):
        data_first = -1
        text_first = -1
        data_seg = False
        text_seg = True
        lines = code.split("\n")
        data_seg_count = 0
        data_seg_found = False
        text_seg_found = False
        for index in range(len(lines)):
            line = lines[index]
            if line.find(';') > -1:
                line = line[:line.find(';')]

            line = line.strip()
            if len(line) == 0:
                continue

            if line[:5] == ".data":
                data_seg = True
                text_seg = False
                data_seg_count += 1
                data_seg_found = True
                continue
            elif line[:5] == ".text":
                data_seg = False
                text_seg = True
                text_seg_found = True
                continue
            if data_seg and data_first == -1:
                data_first = index
            elif text_seg and text_first == -1:
                text_first = index

        if data_seg_found and not text_seg_found:
            return ""
        errors = []
        if data_first > text_first:
            errors.append("Data segment is after text segment.")
        if data_seg_count > 1:
            errors.append("Code has more than one data segment.")
        return "\n".join(errors)

    def parse_changes(self):
        reg_changes = []
        mem_changes = []
        flag_changes = []
        for reg in self.vm_machine.registers:
            if reg in self.vm_machine.changes:
                reg_val = self.vm_machine.registers[reg]
                reg_changes.append((reg, reg_val))

        for chng in self.vm_machine.changes:
            if "MEM" in chng:
                loc = chng.replace("MEM", "")
                mem_changes.append((loc, self.vm_machine.memory[loc]))
            if "FLAG" in chng:
                flag_nm = chng.replace("FLAG", "")
                flag_changes.append((flag_nm, self.vm_machine.flags[flag_nm]))

        return reg_changes, mem_changes, flag_changes

    def construct_table(self, reg_changes, mem_changes, flag_changes):
        output = '<table><tr><th>Type</th><th>Identifier</th><th>Value</th>'
        for reg_name, reg_val in reg_changes:
            output += f'''<tr><td>Register</td><td>{reg_name}</td>
            <td>{reg_val}</td></tr>'''
        for mem_loc, mem_val in mem_changes:
            output += f'''<tr><td>Memory</td><td>{mem_loc}</td>
            <td>{mem_val}</td></tr>'''
        for flag_nm, flag_val in flag_changes:
            output += f'''<tr><td>Flag</td><td>{flag_nm}</td>
            <td>{flag_val}</td></tr>'''
        output += "</table>"
        return output

    def do_execute(self, code, silent, store_history=True,
                   user_expressions=None, allow_stdin=False):

        if not silent:
            if not self.vm_machine:
                self.vm_machine = MIPSMachine()
                self.vm_machine.base = 'hex'
                self.vm_machine.flavor = 'mips_asm'

            self.vm_machine.changes_init()

            data_check_errors = self.has_one_data_first(code)
            if data_check_errors:
                error_msg = {'name': 'error_msg',
                             'text': data_check_errors}
                self.send_response(self.iopub_socket, 'stream', error_msg)

            else:
                (last_instr, error, bit_code) = assemble(code, self.vm_machine,
                                                         web=False)
                if error == "":
                    vm_machine_info = {}
                    reg_info, mem_info, flag_info = self.parse_changes()
                    output = "Changes: <br />"
                    output += self.construct_table(reg_info, mem_info,
                                                   flag_info)
                    vm_machine_info['data'] = {
                        'text/html': output
                    }
                    self.send_response(self.iopub_socket,
                                       'display_data', vm_machine_info)
                else:
                    error_msg = {'name': 'error_msg', 'text': error}
                    self.send_response(self.iopub_socket, 'stream', error_msg)

        return {'status': 'ok',
                'execution_count': self.execution_count,
                'payload': [],
                'user_expressions': {}
                }



import re

"""
Possible assembler errors.
"""

# Error messages (or at least their beginnings):
INT_OUT_OF_RNG = "Integer out of range: "
INVALID_INSTR = "Invalid instruction: "
INVALID_OPRND = "Invalid operand: "
INVALID_NUM_ARGS = "Invalid number of args: "
INVALID_MEM_LOC = "Invalid memory location: "
INVALID_REG = "Invalid register: "
INVALID_VAR_DECL = "Invalid variable declaration: "
INVALID_DATA_TYPE = "Invalid data type: "
INVALID_DATA_VAL = "Invalid data value: "
INVALID_CONSTANT_VAL = "Invalid constant value: "
INVALID_SECTION = "Invalid section: "
INVALID_TOKEN = "Invalid argument: "
INVALID_PC = "Invalid PC: "
LABEL_NOT_SETTABLE = "Label values can't be reset."
NOT_SETTABLE = "This operand type can't have its value set: "
PROGRAM_EXIT = "Program exit"
REG_UNWRITABLE = "Write attempt to unwriteable register: "
STACK_OVERFLOW = "Stack overflow."
STACK_UNDERFLOW = "Stack underflow."
UNKNOWN_ERR = "Unknown parsing error."
UNKNOWN_INT = "Unknown interrupt instruction: "
UNKNOWN_NM = "Unknown symbol: "
UNKNOWN_LABEL = "Unknown label: "
MISSING_OPS = "Missing operands where expected"
MISSING_PC = "Missing PC where expected"
MISSING_DATA = "Missing data values"
MISSING_COMMA = "Missing comma"
MISSING_OPENBRACK = "Missing opening bracket"
MISSING_OPENPAREN = "Missing opening parenthesis"
MISSING_CLOSEBRACK = "Missing closing bracket"
MISSING_CLOSEPAREN = "Missing closing parenthesis"
ZERO_DIVISION = "Division by zero"
OUT_OF_BOUNDS = "Displacement out of bounds"
TOO_BIG_FOR_SINGLE = "Value too big to store in a single precision: "
TOO_BIG_FOR_DOUBLE = "Value too big to store in a double: "
NOT_EVEN_REGISTER = "Invalid odd-numbered register: "
NOT_CORRECT_BITS = "Instruction expected integer of max length 20 bits"
TOO_PRECISE = "Floating point number has too many decimal places: "
INVALID_STRING = "The String Provided is Invalid"
STACK_FULL = "Cannot push another element, stack is full"

INT_MAX = (2**31)-1
INT_MIN = -(2**31)


class Error(Exception):
    """
    Base class for all of our error exceptions.
    """
    def __init__(self, offender, line_num=None):
        self.msg = UNKNOWN_ERR
        self.line_num = line_num


class UnknownInt(Error):
    def __init__(self, offender, line_num):
        self.msg = UNKNOWN_INT + offender
        self.line_num = line_num


class RegUnwritable(Error):
    def __init__(self, offender, line_num):
        self.msg = REG_UNWRITABLE + offender
        self.line_num = line_num


class InvalidVarDeclaration(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_VAR_DECL + offender
        self.line_num = line_num


class InvalidDataType(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_DATA_TYPE + offender
        self.line_num = line_num


class InvalidDataVal(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_DATA_VAL + offender
        self.line_num = line_num


class InvalidConVal(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_CONSTANT_VAL + offender
        self.line_num = line_num


class InvalidInstruction(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_INSTR + offender
        self.line_num = line_num


class InvalidSection(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_SECTION + offender
        self.line_num = line_num


class InvalidArgument(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_TOKEN + offender
        self.line_num = line_num


class InvalidPc(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_PC + offender
        self.line_num = line_num


class NotSettable(Error):
    def __init__(self, offender, line_num):
        self.msg = NOT_SETTABLE + offender
        self.line_num = line_num


class LabelNotSettable(Error):
    def __init__(self, offender, line_num):
        self.msg = LABEL_NOT_SETTABLE + offender
        self.line_num = line_num


class UnknownLabel(Error):
    def __init__(self, offender, line_num):
        self.msg = UNKNOWN_LABEL + offender
        self.line_num = line_num


class UnknownName(Error):
    def __init__(self, offender, line_num):
        self.msg = UNKNOWN_NM + offender
        self.line_num = line_num


class InvalidOperand(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_OPRND + offender
        self.line_num = line_num


class InvalidNumArgs(Error):
    def __init__(self, offender, line_num, correct_num, actual_num,
                 extra_arg=None):
        extra = ""
        if extra_arg is not None:
            extra = "; possible extra = " + str(extra_arg)
        self.msg = (INVALID_NUM_ARGS + offender
                    + " requires " + str(correct_num)
                    + " but we got " + str(actual_num)
                    + extra)
        self.line_num = line_num


class InvalidMemLoc(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_MEM_LOC + offender
        self.line_num = line_num


class InvalidRegister(Error):
    def __init__(self, offender, line_num):
        self.msg = INVALID_REG + offender
        self.line_num = line_num


class IntOutOfRng(Error):
    def __init__(self, offender, line_num=None):
        self.msg = INT_OUT_OF_RNG + offender
        self.line_num = line_num


class StackOverflow(Error):
    def __init__(self, line_num):
        self.msg = STACK_OVERFLOW
        self.line_num = line_num


class StackUnderflow(Error):
    def __init__(self, line_num):
        self.msg = STACK_UNDERFLOW
        self.line_num = line_num


class ExitProg(Error):
    def __init__(self, offender):
        self.msg = offender + ": " + PROGRAM_EXIT


class MissingOps(Error):
    def __init__(self, line_num):
        self.msg = MISSING_OPS
        self.line_num = line_num


class MissingPc(Error):
    def __init__(self, line_num):
        self.msg = MISSING_PC
        self.line_num = line_num


class MissingData(Error):
    def __init__(self, line_num):
        self.msg = MISSING_DATA
        self.line_num = line_num


class MissingComma(Error):
    def __init__(self, line_num):
        self.msg = MISSING_COMMA
        self.line_num = line_num


class MissingOpenParen(Error):
    def __init__(self, line_num):
        self.msg = MISSING_OPENPAREN
        self.line_num = line_num


class MissingOpenBrack(Error):
    def __init__(self, line_num):
        self.msg = MISSING_OPENBRACK
        self.line_num = line_num


class MissingCloseParen(Error):
    def __init__(self, line_num):
        self.msg = MISSING_CLOSEPAREN
        self.line_num = line_num


class MissingCloseBrack(Error):
    def __init__(self, line_num):
        self.msg = MISSING_CLOSEBRACK
        self.line_num = line_num


class DivisionZero(Error):
    def __init__(self, line_num):
        self.msg = ZERO_DIVISION
        self.line_num = line_num


class OutofBounds(Error):
    def __init__(self, line_num):
        self.msg = OUT_OF_BOUNDS
        self.line_num = line_num


class TooBigForSingle(Error):
    def __init__(self, offender, line_num):
        self.msg = TOO_BIG_FOR_SINGLE + offender
        self.line_num = line_num


class TooBigForDouble(Error):
    def __init__(self, offender, line_num):
        self.msg = TOO_BIG_FOR_DOUBLE + offender
        self.line_num = line_num


class NotEvenRegister(Error):
    def __init__(self, offender, line_num):
        self.msg = NOT_EVEN_REGISTER + offender
        self.line_num = line_num


class IncorrectImmLength(Error):
    def __init__(self, offender, line_num):
        self.msg = NOT_CORRECT_BITS + offender
        self.line_num = line_num


class TooPrecise(Error):
    def __init__(self, offender, line_num):
        self.msg = TOO_PRECISE + offender
        self.line_num = line_num


class StackFull(Error):
    def __init__(self, line_num):
        self.msg = STACK_FULL
        self.line_num = line_num


def check_num_args(instr, ops, correct_num, line_num, type_ins=0):
    """
    See if we have the proper number of arguments.
    """
    length = len(ops)
    if length != correct_num:
        extra_arg = None
        if length > correct_num:
            extra_arg = ops[length - 1]
        raise InvalidNumArgs(instr, line_num, correct_num, len(ops),
                             extra_arg)
    for i in range(0, length):
        if type_ins == 0 and (ops[i].get_val(line_num) > INT_MAX or
                              ops[i].get_val(line_num) < INT_MIN):
            raise IntOutOfRng(instr, line_num)


def is_notvalidstring(instr):
    match = re.match("^(\d{0,3})\.(\d{0,3})\.(\d{0,3})\.(\d{0,3})$", instr) # noqa!



"""
fp_arithmetic.py: arithmetic and logic instructions.
"""

import operator as opfunc

from assembler.errors import check_num_args, TooBigForSingle, DivisionZero
from assembler.tokens import Instruction, MAX_INT
from .argument_check import check_reg_only, check_even_register

# for floating point to binary and back
import struct
import binascii


def check_overflow(val, vm):
    if(val > MAX_INT):
        val = val - MAX_INT+1
    return val


def three_op_arith_reg(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num, type_ins=1)
    check_reg_only(instr, ops, line_num)

    # go through the register ops and make sure that they're even numbered
    for op in ops:
        check_even_register(op, line_num)

    ops[0].set_val(operator(ops[1].get_val(line_num),
                   ops[2].get_val(line_num)), line_num)
    # check_overflow(operator(ops[1].get_val(line_num),
    #                    ops[2].get_val(line_num)),
    #                    vm))
    vm.changes.add(ops[0].get_nm())


# to convert a float to a hex
# using double for a significant amount of precisions
# (i think its up to 48 bits of precision)
def float_to_hex(f):
    return hex(struct.unpack('<I', struct.pack('<f', f))[0])


# to convert the ieee 754 hex back to the actual float value
def hex_to_float(h):
    h2 = h[2:]
    h2 = binascii.unhexlify(h2)
    return struct.unpack('>f', h2)[0]


# 'ADD.S': Adds('ADD.S'),
class Adds(Instruction):
    """
        <instr>
             ADD.S
        </instr>
        <syntax>
            ADD.S reg, reg, reg
        </syntax>
        <descr>
            The 32 bit single precision floating point value in register Fs is
            added to the 32 bit single precision floating point value in
            register Ft to produce a 32 bit single precision floating point
            result. The result is stored in register Fd.
        </descr>
    """
    # ops is a list of operands (reg, reg, reg)
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.add, line_num)


# 'SUB.S': Subs('SUB.S'),
class Subs(Instruction):
    """
        <instr>
             SUB.S
        </instr>
        <syntax>
            SUB.S reg, reg, reg
        </syntax>
        <descr>
            The 32 bit single precision floating point value in register Fs is
            subtracted from the 32 bit single precision floating point value in
            register Ft to produce a 32 bit single precision floating point
            result. The result is stored in register Fd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.sub, line_num)


# 'MULT.S': Mults('MULT.S'),
class Mults(Instruction):
    """
        <instr>
             MULT.S
        </instr>
        <syntax>
            MULT.S reg, reg, reg
        </syntax>
        <descr>
            The 32 bit single precision floating point value in register Fs is
            multiplied by the 32 bit single precision floating point value in
            register Ft to produce a 32 bit single precision floating point
            result. The result is stored in register Fd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num, type_ins=1)
        check_reg_only(self.name, ops, line_num)
        a = ops[1].get_val(line_num)
        b = ops[2].get_val(line_num)
        result = a * b
        # this is the single version of mult for floats
        # so we don't want to be bigger than max single
        if (result > 2 ** 22):
            raise TooBigForSingle(str(result), line_num)

        ops[0].set_val(result, line_num)
        vm.changes.add(ops[0].get_nm())
        return ''


# 'DIV.S': Divs('DIV.S'),
class Divs(Instruction):
    """
        <instr>
             DIVS
        </instr>
        <syntax>
            DIVS reg, reg, reg
        </syntax>
        <descr>
            the 32 bit single precision floating point value in register Fs is
            divided by the 32 bit single precision floating point value in
            register Ft to produce a 32 bit single precision floating point
            result. The result is stored in register Fd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num, type_ins=1)
        check_reg_only(self.name, ops, line_num)
        if ops[2].get_val(line_num) == 0:
            raise DivisionZero(line_num)

        result = ops[1].get_val(line_num) / ops[2].get_val(line_num)
        # this is the single version of div for floats
        # so we don't want to be bigger than max single
        if (result > 2 ** 22):
            raise TooBigForSingle(str(result), line_num)
        ops[0].set_val(result, line_num)
        vm.changes.add(ops[0].get_nm())
        return ''

########################
# DOUBLE PRECISION BELOW
########################
# for double precision (64 bits) fps
# references:
# https://docs.python.org/2/library/struct.html
# https://stackoverflow.com/questions/52600983/converting-float-to-ieee754
# https://stackoverflow.com/questions/19414847/how-to-convert-floating-point-number-in-python # noqa
getBin = lambda x: x > 0 and str(bin(x))[2:] or "-" + str(bin(x))[3:] # noqa


def f_to_b64(value):
    """

    :param value: float
    :return: binary 64 conversion of value in string format
    """
    if (value == 0):
        return "0"*64
    val = struct.unpack('q', struct.pack('d', value))[0]
    return "0" + getBin(val)


def b_to_f64(value):
    """

    :param value: binary value in string format
    :return: float 64 of binary number
    """
    if (value == "0"*64):
        return 0.0
    hx = hex(int(value, 2))
    return struct.unpack("d", struct.pack("q", int(hx, 16)))[0]


def three_op_double_arith_reg(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num, type_ins=1)
    check_reg_only(instr, ops, line_num)

    is_a_neg = False
    is_b_neg = False

    # go through the register ops and make sure that they're even numbered
    # they are even good
    # first number from the pair of registers
    reg_number = int(ops[1].get_nm()[1:])
    curr_reg = "F" + str(reg_number + 0)
    next_reg = "F" + str(reg_number + 1)
    a_first_32 = vm.registers[curr_reg]
    a_last_32 = vm.registers[next_reg]
    a = a_first_32 + a_last_32
    if a != 0 and a[0] == "1":
        is_a_neg = True
        a = "0" + a[1:]

    # second number from the pair of registers
    reg_number = int(ops[2].get_nm()[1:])
    curr_reg = "F" + str(reg_number + 0)
    next_reg = "F" + str(reg_number + 1)
    b_first_32 = vm.registers[curr_reg]
    b_last_32 = vm.registers[next_reg]
    b = b_first_32 + b_last_32
    if b != 0 and b[0] == "1":
        is_b_neg = True
        b = "0" + b[1:]

    if a != 0:
        a = b_to_f64(a)

    if b != 0:
        b = b_to_f64(b)

    if is_a_neg:
        a *= -1
    if is_b_neg:
        b *= -1

    res = operator(a, b)
    is_res_neg = False
    if res < 0:
        is_res_neg = True
        res = abs(res)

    res_binary = f_to_b64(res)

    if (is_res_neg):
        res_binary = "1" + res_binary[1:]

    res_first_32 = res_binary[:32]
    res_last_32 = res_binary[32:]

    # save the result

    reg_number = int(ops[0].get_nm()[1:])
    curr_reg = "F" + str(reg_number + 0)
    next_reg = "F" + str(reg_number + 1)
    ops[0].set_val(res_first_32, line_num)
    vm.registers[next_reg] = res_last_32

    vm.changes.add(ops[0].get_nm())
    # vm.changes.add(vm.registers[curr_reg])
    vm.changes.add(vm.registers[next_reg])


class Addd(Instruction):
    """
        <instr>
             ADD.D
        </instr>
        <syntax>
            ADD.D reg, reg, reg
        </syntax>
        <descr>
            The 64 bit double precision floating point value in registers Fs
            and Fs + 1 are added to the 64 bit double precision floating point
            value in registers Ft and Ft + 1 to get a 64 bit double precision
            floating point result. The result is stored in registers Fd and
            Fd + 1.
        </descr>
    """
    # ops is a list of operands (reg, reg, reg)
    def fhook(self, ops, vm, line_num):
        three_op_double_arith_reg(ops, vm, self.name, opfunc.add, line_num)


class Subd(Instruction):
    """
        <instr>
            SUB.D
        </instr>
        <syntax>
            SUB.D reg, reg, reg
        </syntax>
        <descr>
            The 64 bit double precision floating point value in registers Fs
            and Fs + 1 are subtracted from the 64 bit double precision floating
            point value in registers Ft and Ft + 1 to get a 64 bit double
            precision floating point result. The result is stored in registers
            Fd and Fd + 1.
        </descr>
    """
    # ops is a list of operands (reg, reg, reg)
    def fhook(self, ops, vm, line_num):
        three_op_double_arith_reg(ops, vm, self.name, opfunc.sub, line_num)


class Multd(Instruction):
    """
        <instr>
            MULT.D
        </instr>
        <syntax>
            MULT.D reg, reg, reg
        </syntax>
        <descr>
            The 64 bit double precision floating point value in registers Fs
            and Fs + 1 are multiplied by the 64 bit double precision floating
            point value in registers Ft and Ft + 1 to get a 64 bit double
            precision floating point result. The result is stored in registers
            Fd and Fd + 1.
        </descr>
    """
    # ops is a list of operands (reg, reg, reg)
    def fhook(self, ops, vm, line_num):
        three_op_double_arith_reg(ops, vm, self.name, opfunc.mul, line_num)


class Divd(Instruction):
    """
        <instr>
            DIV.D
        </instr>
        <syntax>
            DIV.D reg, reg, reg
        </syntax>
        <descr>
            The 64 bit double precision floating point value in regsiters Fs
            and Fs + 1 are multiplied by the 64 bit double precision floating
            point value in registers Ft and Ft + 1 to get a 64 bit double
            precision floating point result. The result is stored in registers
            Fd and Fd + 1.
        </descr>
    """
    # ops is a list of operands (reg, reg, reg)
    def fhook(self, ops, vm, line_num):
        three_op_double_arith_reg(ops, vm, self.name, opfunc.truediv, line_num)



"""
arithmetic.py: arithmetic and logic instructions.
"""

import operator as opfunc
from .argument_check import check_reg_only, check_immediate_three
from assembler.errors import check_num_args, DivisionZero, InvalidArgument

from assembler.tokens import Instruction, MAX_INT, Register


def three_op_arith_reg(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    ops[0].set_val(
        check_overflow(
            operator(ops[1].get_val(line_num),
                     ops[2].get_val(line_num)),
            vm),
        line_num)
    vm.changes.add(ops[0].get_nm())


def three_op_arith_immediate(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    ops[0].set_val(
        check_overflow(
            operator(ops[1].get_val(line_num),
                     ops[2].get_val(line_num)),
            vm),
        line_num)
    vm.changes.add(ops[0].get_nm())


def check_overflow(val, vm):
    if(val > MAX_INT):
        val = val - MAX_INT+1
    return val


class Add(Instruction):
    """
        <instr>
             ADD
        </instr>
        <syntax>
            ADD reg, reg, reg
        </syntax>
        <descr>
            The 32 bit word value in Rt is added to the 32 bit word value in Rs
            to produce a 32 bit result. If the addition results in 32 bit 2's
            complement arithmetic overflow, the destination register is not
            modified and an Integer Overflow exception occurs. If the addition
            deos not overflow, the 32 bit result is placed into Rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.add, line_num)


class Addi(Instruction):
    """
        <instr>
             ADDI
        </instr>
        <syntax>
            ADDI reg, reg, con
        </syntax>
        <descr>
            The 16 bit signed immediate is added to the 32 bit value in Rs to
            produce a 32 bit result. If the addition in 32 bit 2's complement
            arithmetic overflow, the destination register is not modified and
            an Integer Overflow exception occurs. If the addition does not
            overflow, the 32 bit result is placed into Rt.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.add, line_num)


class Sub(Instruction):
    """
        <instr>
             SUB
        </instr>
        <syntax>
            SUB reg, reg, reg
        </syntax>
        <descr>
            The 32 bit word value in Rt is subtracted from the 32 bit value in
            Rs to produce a 32 bit result. If the subtraction results in a 32
            bit 2's complement arithmetic overflow, then the destination
            register is not modified and an Integer Overflow exception occurs.
            If it does not overflow, the 32 bit result is placed into Rd.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.sub, line_num)


class Mult(Instruction):
    """
        <instr>
             MULT
        </instr>
        <syntax>
            MULT reg, reg
        </syntax>
        <descr>
            The 32 bit word value in Rt is multiplied by the 32 bit value in Rs
            to produce a 64 bit result. The upper 32 bits of the 64 bit result
            is placed in the special register HI, whereas the lower 32 bits are
            placed in the register LO. Use MFHI and MFLO to move the result to
            a general purpose register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 2, line_num)
        check_reg_only(self.name, ops, line_num)
        result = ops[0].get_val(line_num) * ops[1].get_val(line_num)
        if result > 2 ** 32 - 1:
            vm.registers['LO'] = opfunc.lshift(result, 32)
            vm.registers['HI'] = opfunc.rshift(result, 32)
        else:
            vm.registers['LO'] = result
            vm.registers['HI'] = 0
        vm.changes.add('LO')
        vm.changes.add('HI')
        return ''


class Andf(Instruction):
    """
        <instr>
             AND
        </instr>
        <syntax>
            AND reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise AND operation on registers Rt and Rs and stores
            the result in register Rd. If the bit in both Rt and Rs is 1, Rd
            will get 1; otherwise it will get 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.and_, line_num)
        return ''


class Andi(Instruction):
    """
        <instr>
             AND
        </instr>
        <syntax>
            ANDI reg, reg, con
        </syntax>
        <descr>
            Performs a bitwise AND operation on register Rs and a 16 bit sign-
            extended immediate extended to 32 bits and stores the result in
            register Rt. If the bit in both Rs and the immediate is 1, Rt gets
            a 1; otherwise it will get a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.and_, line_num)
        return ''


class Orf(Instruction):
    """
        <instr>
             OR
        </instr>
        <syntax>
            OR reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise OR operation on registers Rs and Rt and stores
            the result in register Rd. If the bit in both Rs and Rt is 0, Rd
            gets a 0; otherwise it will get a 1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.or_, line_num)
        return ''


class Ori(Instruction):
    """
        <instr>
             OR
        </instr>
        <syntax>
            ORI reg, reg, con
        </syntax>
        <descr>
            Performs a bitwise OR operation on register Rs and a 16 bit sign-
            extended immediate extended to 32 bits and stores the result in Rt.
            If the bit in both Rs and the immediate is 0, Rt gets a 0;
            otherwise it will get a 1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.or_, line_num)
        return ''


class Nor(Instruction):
    """
        <instr>
             NOR
        </instr>
        <syntax>
            NOR reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise NOR operation on registers Rs and Rt and stores
            the result in register Rd. If the bit in both Rs and Rt is 0, Rd
            gets a 1; otherwise it will get a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.or_, line_num)
        ops[0].set_val(opfunc.inv(ops[0].get_val(line_num)), line_num)
        return ''


class Xor(Instruction):
    """
        <instr>
             XOR
        </instr>
        <syntax>
            XOR reg, reg, reg
        </syntax>
        <descr>
            Performs a bitwise XOR operation on registers Rs and Rt and stores
            the result in register Rd. If the bit in Rs and Rt is different, Rd
            gets a 1; otherwise it will get a 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.xor, line_num)
        return ''


class Sll(Instruction):
    """
        <instr>
             sll
        </instr>
        <syntax>
            SLL reg, reg, con
        </syntax>
        <descr>
            The 32 bit word value in register Rs is shifted to the left by a
            number of bits specified by the 16 bit immediate. The result is
            stored as a 32 bit word value in register Rt. This is equivalent
            to multiplying by a power of 2. Bits shifted beyond the boundary of
            Rt are discarded.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.lshift, line_num)
        return ''


class Srl(Instruction):
    """
        <instr>
             srl
        </instr>
        <syntax>
            SRL reg, reg, con
        </syntax>
        <descr>
            The 32 bit word value in register Rs is shifted to the right by a
            number of bits specified by the 16 bit immediate. The result is
            stored as a 32 bit word value in register Rt. This is equivalent to
            dividing by a power of 2. Bits shifted beyond the boundary of Rt
            are discarded.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.rshift, line_num)


class Div(Instruction):
    """
        <instr>
             DIV
        </instr>
        <syntax>
            DIV reg, reg
        </syntax>
        <descr>
            The 32 bit word value in register Rs is divided by the 32 bit word
            value in register Rt. The 32 bit result of the division is stored
            in special register LO. The 32 bit modulus of the division is
            stored in the special register HI. Use MFHI and MFLO to move these
            results to general purpose registers.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 2, line_num)
        check_reg_only(self.name, ops, line_num)
        if ops[1].get_val(line_num) == 0:
            raise DivisionZero(line_num)

        quotient = ops[0].get_val(line_num) // ops[1].get_val(line_num)
        remainder = ops[0].get_val(line_num) % ops[1].get_val(line_num)
        vm.registers['LO'] = quotient
        vm.registers['HI'] = remainder
        vm.changes.add('LO')
        vm.changes.add('HI')
        return ''


class Mfhi(Instruction):
    """
        <instr>
             mfhi
        </instr>
        <syntax>
            MFHI reg
        </syntax>
        <descr>
            Moves the value from the HI register into the
            destination register given.
        </descr>
    """

    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidArgument(ops[0].get_nm(), line_num)
        ops[0].set_val(vm.registers['HI'], line_num)
        vm.changes.add(ops[0].get_nm())
        return ''


class Mflo(Instruction):
    """
        <instr>
             mflo
        </instr>
        <syntax>
            MFLO reg
        </syntax>
        <descr>
            Moves the value from the LO register into the
            destination register given.
        </descr>
    """

    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidArgument(ops[0].get_nm(), line_num)
        ops[0].set_val(vm.registers['LO'], line_num)
        vm.changes.add(ops[0].get_nm())
        return ''



from assembler.tokens import Register, IntegerTok, Symbol
from assembler.errors import InvalidArgument


def check_reg_only(instr, ops, line_num):
    for i in range(0, len(ops)):
        if not isinstance(ops[i], Register):
            raise InvalidArgument(ops[i].get_nm(), line_num)


def check_immediate_three(instr, ops, line_num):
    if not isinstance(ops[0], Register):
        raise InvalidArgument(ops[0].get_nm(), line_num)
    else:
        if isinstance(ops[1], Register):
            if (not isinstance(ops[2], IntegerTok) and
                    not isinstance(ops[2], Symbol)):
                raise InvalidArgument(ops[2].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[1].get_nm(), line_num)


def check_immediate_two(instr, ops, line_num):
    if isinstance(ops[0], Register):
        if not isinstance(ops[1], IntegerTok):
            raise InvalidArgument(ops[1].get_nm(), line_num)
    else:
        raise InvalidArgument(ops[0].get_nm(), line_num)



from .arithmetic import Add, Addi, Sub, Mul, And, Andi
from .arithmetic import Xor, Xori, Or, Ori, Srl, Sll
from .arithmetic import Srli, Slli, Slt, Sltu, Slti, Sltiu
from .arithmetic import Sra, Srai, Div, Divu, Rem, Remu
from .arithmetic import Lui
from .control_flow import Jr, Jal, Jalr, Beq, Bne, Blt
from .data_mov import Load, Store
from .interrupts import Syscall
from assembler.tokens import DataType

key_words = {
    # data types
    '.BYTE': DataType('DB'),
    '.WORD': DataType('DW'),

    # data movement:
    'LW': Load('LW'),
    'SW': Store('SW'),

    # arithmetic and logic
    'ADD': Add('ADD'),
    'ADDI': Addi('ADDI'),
    'SUB': Sub('SUB'),
    'MUL': Mul('MUL'),
    'AND': And('AND'),
    'ANDI': Andi('ANDI'),
    'OR': Or('OR'),
    'ORI': Ori('ORI'),
    'XOR': Xor('XOR'),
    'XORI': Xori('XORI'),
    'SRL': Srl('SRL'),
    'SLL': Sll('SLL'),
    'SLLI': Slli('SLLI'),
    'SRLI': Srli('SRLI'),
    'SLT': Slt('SLT'),
    'SLTU': Sltu('SLTU'),
    'SLTI': Slti('SLTI'),
    'SLTIU': Sltiu('SLTIU'),
    'SRA': Sra('SRA'),
    'SRAI': Srai('SRAI'),
    'DIV': Div('DIV'),
    'DIVU': Divu('DIVU'),
    'REM': Rem('REM'),
    'REMU': Remu('REMU'),
    'LUI': Lui('LUI'),

    # control
    'JR': Jr('Jr'),
    'JAL': Jal('JAL'),
    'JALR': Jalr('JALR'),
    'BEQ': Beq('BEQ'),
    'BNE': Bne('BNE'),
    'BLT': Blt('BLT'),

    # interrupts
    'SYSCALL': Syscall('SYSCALL'),
}

op_func_codes = {
    # R-format
    'ADD': '0110011',
    'SUB': '0110011',
    'MUL': '0110011',
    'AND': '0110011',
    'OR': '0110011',
    'XOR': '0110011',
    'SRL': '0110011',
    'SLL': '0110111',
    'SLT': '0110011',
    'SLTU': '0110011',
    'SRA': '0110011',
    'DIV': '0110011',
    'DIVU': '0110011',
    'REM': '0110011',
    'REMU': '0110011',

    # I-format
    'LW': '0000011',
    'ADDI': '0010011',
    'ANDI': '0010011',
    'ORI': '0010011',
    'XORI': '0010011',
    'SRLI': '0010011',
    'SLLI': '0010011',
    'SLTI': '0010011',
    'SLTIU': '0010011',
    'SRAI': '0010011',
    'LUI': '0110111',
    'SYSCALL': '0001100',

    # S-Format
    'SW': '0100011',

    # J-format
}



from setuptools import setup

with open('Emu86KernelPYPIDescription.md', 'r') as f:
    long_description = f.read()

setup(
    name='emu86',
    version='0.0.10',
    packages=['kernels', 'kernels/att', 'kernels/intel', 'kernels/riscv',
              'kernels/mips_asm', 'kernels/mips_mml',
              'assembler', 'assembler/Intel', 'assembler/MIPS',
              'assembler/RISCV', 'assembler/WASM'],
    long_description=long_description,
    long_description_content_type='text/markdown',
)



from assembler.errors import check_num_args, InvalidArgument
from assembler.tokens import Instruction, NewSymbol, IntegerTok


class Global_get(Instruction):
    """
        <instr>
             global.get
        </instr>
        <syntax>
            global.get var
        </syntax>
        <descr>
            Copies the value of op1 onto the stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], NewSymbol):
            if ops[0].get_nm() in vm.globals:
                stack_loc = hex(vm.get_sp()).split('x')[-1].upper()
                vm.stack[stack_loc] = vm.globals[ops[0].get_nm()]
                vm.inc_sp(line_num)
            else:
                raise InvalidArgument(ops[0].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Global_set(Instruction):
    """
        <instr>
             global.set
        </instr>
        <syntax>
            global.set var
        </syntax>
        <descr>
            Copies the value of op1 onto the stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], NewSymbol):
            if ops[0].get_nm() in vm.globals:
                vm.dec_sp(line_num)
                stack_loc = hex(vm.get_sp()).split('x')[-1].upper()
                vm.globals[ops[0].get_nm()] = vm.stack[stack_loc]
                vm.inc_sp(line_num)
                vm.changes.add(f'GLOBALVAR{ops[0].get_nm()}')
            else:
                raise InvalidArgument(ops[0].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Local_get(Instruction):
    """
        <instr>
             local.get
        </instr>
        <syntax>
            local.get var
        </syntax>
        <descr>
            Copies the value of op1 onto the stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], NewSymbol):
            if ops[0].get_nm() in vm.locals:
                stack_loc = hex(vm.get_sp()).split('x')[-1].upper()
                vm.stack[stack_loc] = vm.locals[ops[0].get_nm()]
                vm.inc_sp(line_num)
            else:
                raise InvalidArgument(ops[0].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Local_set(Instruction):
    """
        <instr>
             local.set
        </instr>
        <syntax>
            local.set var
        </syntax>
        <descr>
            Copies the value of op1 onto the stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], NewSymbol):
            if ops[0].get_nm() in vm.locals:
                vm.dec_sp(line_num)
                stack_loc = hex(vm.get_sp()).split('x')[-1].upper()
                vm.locals[ops[0].get_nm()] = vm.stack[stack_loc]
                vm.inc_sp(line_num)
                vm.changes.add(f'LOCALVAR{ops[0].get_nm()}')
            else:
                raise InvalidArgument(ops[0].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Store_global(Instruction):
    """
        <instr>
             global
        </instr>
        <syntax>
            global var
        </syntax>
        <descr>
            Store a global value into the globals dictionary
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], NewSymbol):
            vm.globals[ops[0].get_nm()] = ops[0].get_val(line_num)
            vm.changes.add(f'GLOBALVAR{ops[0].get_nm()}')
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Store_local(Instruction):
    """
        <instr>
             local
        </instr>
        <syntax>
            local var
        </syntax>
        <descr>
            Store a local value into the locals dictionary
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], NewSymbol):
            vm.locals[ops[0].get_nm()] = ops[0].get_val(line_num)
            vm.changes.add(f'LOCALVAR{ops[0].get_nm()}')
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Store_const(Instruction):
    """
        <instr>
             i32.const
        </instr>
        <syntax>
            i32.const val
        </syntax>
        <descr>
            Store a constant value onto the stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if isinstance(ops[0], IntegerTok):
            try:
                stack_loc = hex(vm.get_sp()).split('x')[-1].upper()
                vm.stack[stack_loc] = ops[0].get_val(line_num)
                vm.inc_sp(line_num)
            except Exception:
                raise InvalidArgument(ops[0].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)



"""
arithmetic.py: arithmetic and logic instructions.
"""

import operator as opfunc

from assembler.errors import DivisionZero, check_num_args, InvalidConVal
from assembler.errors import InvalidOperand
from assembler.tokens import Instruction, MAX_INT, Register
from assembler.ops_check import one_op_arith


def two_op_arith(ops, vm, instr, line_num, operator):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 2, line_num)
    ops[0].set_val(
        checkflag(operator(ops[0].get_val(line_num),
                           ops[1].get_val(line_num)), vm), line_num)
    vm.changes.add(ops[0].get_nm())


def checkflag(val, vm):
    if(val > MAX_INT):
        vm.flags['CF'] = 1
        val = val - MAX_INT+1
    else:
        vm.flags['CF'] = 0
    return val


class Add(Instruction):
    """
        <instr>
             add
        </instr>
        <syntax>
            ADD reg, reg
            ADD reg, mem
            ADD reg, con
        </syntax>
        <descr>
            Adds the first operand with the second and stores the result in
            the first operand. The destination can be a register or memory
            location; the source can be an immediate, register, or memory
            location. Two memory operands cannot be used. Immediate values are
            sign extended. Works on both signed and unsigned integer operands.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.add)


class Sub(Instruction):
    """
        <instr>
             sub
        </instr>
        <syntax>
            SUB reg, reg
            SUB reg, mem
            SUB reg, con
        </syntax>
        <descr>
            Subtracts the second operand from the first operand and stores the
            result in the first operand. The destination can be a register or
            memory location; the source can be an immediate, register, or
            memory location. Two memory operands cannot be used. Immediate
            values are sign extended. Evaluates the result for both signed
            and unsigned integer operands.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.sub)


class Imul(Instruction):
    """
        <instr>
             imul
        </instr>
        <syntax>
            IMUL reg, reg
            IMUL reg, mem
            IMUL reg, con
        </syntax>
        <descr>
            Performs a signed multiplication of two operands. The first
            operand is multiplied by the second operand. The destination is a
            general purpose register and the source can be an immediate, a
            register, or a memory location. The product is truncated and
            stored in the destination operand location.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.mul)
        return ''


class Andf(Instruction):
    """
        <instr>
             and
        </instr>
        <syntax>
            AND reg, reg
            AND reg, mem
            AND reg, con
        </syntax>
        <descr>
            Performs a bitwise AND operation on the destination (first) and
            source (second) operands and stores the result in the destination
            operand location. The source operand can be an immediate, a
            register, or a memory location; the destination operand can be a
            register or memory location. Two memory operands cannot be used in
            one instruction. Each bit of the result is set to 1 if both
            corresponding bits of the first and second operands are 1;
            otherwise it is set to 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.and_)
        return ''


class Orf(Instruction):
    """
        <instr>
             or
        </instr>
        <syntax>
            OR reg, reg
            OR reg, mem
            OR reg, con
        </syntax>
        <descr>
            Performs a bitwise inclusive OR operation between the destination
            (first) and source (second) operands and stores the result in the
            destination operand location. The source operand can be an
            immediate, a register, or a memory location; the destination
            operand can be a register or a memory location. Two memory
            operands cannot be used in one instruction. Each bit of the result
            of the OR instruction is set to 0 if both corresponding bits of
            the first and second operands are 0; otherwise, each bit is set to
            1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.or_)
        return ''


class Xor(Instruction):
    """
        <instr>
             xor
        </instr>
        <syntax>
            XOR reg, reg
            XOR reg, mem
            XOR reg, con
        </syntax>
        <descr>
            Performs a bitwise exlcusive OR (XOR) operation on the destination
            (first) and source (second) operands and stores the result in the
            destination operand location. The source operand can be an
            immediate, a register, or a memory location; the destination
            operand can be a register or a memory location. Two memory
            operands cannot be used in one instruction. Each bit of the result
            is 1 if the corresponding bits of the operands are different; each
            bit is 0 if the corresponding bits are the same.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.xor)
        return ''


class Shl(Instruction):
    """
        <instr>
             shl
        </instr>
        <syntax>
            SHL reg, reg
            SHL reg, mem
            SHL reg, con
        </syntax>
        <descr>
            Shifts the bits in the first operand (destination operand) to the
            left by the number of bits specified in the second operand
            (count operand). Bits shifted beyonf the destination operand
            boundary are first shifted into the CF flag, then discarded. At the
            end of the shift operation, the CF flag contains the last bit
            shifted out of the destination operand.
            The destination operand can be a register or a memory location. The
            count operand can be an immediate value or the CL register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.lshift)
        return ''


class Shr(Instruction):
    """
        <instr>
             shr
        </instr>
        <syntax>
            SHR reg, reg
            SHR reg, mem
            SHR reg, con
        </syntax>
        <descr>
            Shifts the bits in the first operand (destination operand) to the
            right by the number of bits specified in the second operand
            (count operand). Bits shifted beyond the destination operand
            boundary are first shifted into the CF flag, then discarded. At the
            end of the shift operation, the CF flag contains the last bit
            shifted out of the destination operand.
            The destination operand can be a register or a memory location. The
            count operand can be an immediate value or a register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.rshift)


class Notf(Instruction):
    """
        <instr>
             not
        </instr>
        <syntax>
            NOT reg
        </syntax>
        <descr>
            Performs a bitwise NOT operation (each 1 is set to 0, and each 0 is
            set to 1) on the destination operand and stores the result in the
            destination operand location. The destination operand is a
            register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        one_op_arith(ops, vm, self.name, line_num, opfunc.inv)


class Inc(Instruction):
    """
        <instr>
             inc
        </instr>
        <syntax>
            INC reg
        </syntax>
        <descr>
            Adds 1 to the destination operand, while preserving the state of
            the CF flag. The destination operand must be a register. This
            instruction allows a loop counter to be updated without disturbing
            the CF flag. (Use an ADD instriction with an immediate operand of 1
            to perform an increment operation that does update the CF flag.)
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        ops[0].set_val(ops[0].get_val(line_num) + 1, line_num)
        vm.changes.add(ops[0].get_nm())


class Dec(Instruction):
    """
        <instr>
             dec
        </instr>
        <syntax>
            DEC reg
        </syntax>
        <descr>
            Subtracts 1 from the destination operand, while preserving the
            state of the CF flag. The destination operand must be a register.
            This instruction allows a loop counter to be updated without
            disturbing the CF flag. (To perform a decrement operation that
            updates the CF flag, use a SUB instruction with an immediate
            operand of 1.)
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        ops[0].set_val(ops[0].get_val(line_num) - 1, line_num)
        vm.changes.add(ops[0].get_nm())


class Neg(Instruction):
    """
        <instr>
             neg
        </instr>
        <syntax>
            NEG reg
        </syntax>
        <descr>
           Replaces the value of operand (the destination operand) with its
           two's complement. (This operation is equivalent to subracting the
           operand from 0.) The destination operand is located in a general-
           purpose register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        one_op_arith(ops, vm, self.name, line_num, opfunc.neg)
        return ''


class Idiv(Instruction):
    """
        <instr>
             idiv
        </instr>
        <syntax>
            IDIV reg
        </syntax>
        <descr>
            The idiv instruction divides the contents of the 64 bit integer
            EDX:EAX (constructed by viewing EDX as the most significant four
            bytes and EAX as the least significant four bytes) by the specified
            operand value. The quotient result of the division is stored into
            EAX, while the remainder is placed in EDX.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)

        hireg = int(vm.registers['EDX']) << 32
        lowreg = int(vm.registers['EAX'])
        dividend = hireg + lowreg
        if ops[0].get_val(line_num) == 0:
            raise DivisionZero()
        vm.registers['EAX'] = dividend // ops[0].get_val(line_num)
        vm.registers['EDX'] = dividend % ops[0].get_val(line_num)
        vm.changes.add('EAX')
        vm.changes.add('EDX')
        return ''


class BTR(Instruction):
    """
    <instr>
        btr
    </instr>
    <syntax>
        btr reg, reg
        btr reg, const
    </syntax>
    <descr>
        Selects the bit in a bit string (specified with the first operand,
        called the bit base) at the bit-position designated by the bit offset
        operand (second operand), stores the value of the bit in the CF flag,
        and clears the selected bit in the bit string to 0. The bit base
        operand can be a register or a memory location; the bit offset operand
        can be a register or an immediate value.
        If the bit base operand is a register, the instruction takes the modulo
        16, 32, or 64 of the bit offset operand (modulo size depends on the
        mode and register size). This allows any bit position to be selected.
    </descr>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        if ops[1].get_val(line_num).bit_length() >= 9:
            raise InvalidConVal(ops[1].get_nm(), line_num)
        ops[0].set_val(set_bit(ops[0].get_val(line_num),
                               ops[1].get_val(line_num), 0), line_num)


def set_bit(v, index, x):
    mask = 1 << index   # Compute mask, an integer with just bit 'index' set.
    v &= ~mask          # Clear the bit indicated by the mask (if x is False)
    if x:
        v |= mask         # If x was True, set the bit indicated by the mask.
    return v


class BTS(Instruction):
    """
    <instr>
        bts
    </instr>
    <syntax>
        bts reg, reg
        bts reg, const
    </syntax>
    <descr>
        Selects the bit in a bit string (specified with the first operand,
        called the bit base) at the bit-position designated by the bit offset
        operand (second operand), stores the value of the bit in the CF flag,
        and sets the selected bit in the bit string to 1. The bit base operand
        is a register; the bit offset operand can be a register or an immediate
        value.
        If the bit base operand is a register, the instruction takes the modulo
        16, 32, or 64 of the bit offset operand (modulo size depends on the
        mode and register size.) This allows any bit position to be selected.
    </descr>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        if ops[1].get_val(line_num).bit_length() >= 9:
            raise InvalidConVal(ops[1].get_nm(), line_num)
        ops[0].set_val(set_bit(ops[0].get_val(line_num),
                               ops[1].get_val(line_num), 1), line_num)


class BSF(Instruction):
    """
    <instr>
        bsf
    </instr>
    <syntax>
        bsf reg, reg
        bsf reg, mem
    </syntax>
    <descr>
        Searches the source operand (second operand) for the least significant
        set bit (1 bit). If a least significant 1 bit is found, its bit index
        is stored in the destination operand (first operand). The source
        operand can be a register or memory location; the destination operand
        is a register. The bit index is an unsigned offset from bit 0 of the
        source operand. If the content of the source operand is 0, the content
        of the destination operand is undefined.
    </descr>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        num = ops[1].get_val(line_num)

        if num == 0:
            vmachine.flags["ZF"] = 1
        else:
            vmachine.flags["ZF"] = 0
            if ops[1].get_val(line_num).bit_length() <= 16:
                bit_size = 16
            else:
                bit_size = 32
            bin_str = str(bin(num))[2:].zfill(bit_size)[::-1]
            for index in range(len(bin_str)):
                if bin_str[index] == '1':
                    break
            ops[0].set_val(index, line_num)


class BSR(Instruction):
    """
    <instr>
        bsr
    </instr>
    <syntax>
        bsr reg, reg
        bsr reg, mem
    </syntax>
    <descr>
        Searches the source operand (second operand) for the most significant
        set bit (1 bit). If a most significant 1 bit is found, its bit index is
        stored in the destination operand (first operand). The source operand
        can be a register or a memory location; the destination operand is a
        register. The bit index is an unsigned offset from bit 0 of the source
        operand. If the content source operand is 0, the content of the
        destination operand is undefined.
    </descr>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is notof type Register', line_num)
        num = ops[1].get_val(line_num)
        if num == 0:
            vmachine.flags["ZF"] = 1
        else:
            vmachine.flags["ZF"] = 0
            if ops[1].get_val(line_num).bit_length() <= 16:
                bit_size = 16
            else:
                bit_size = 32
            binStr = str(bin(num))[2:].zfill(bit_size)
            for index in range(len(binStr)):
                if binStr[index] == '1':
                    break
            index = bit_size - index
            ops[0].set_val(index, line_num)


class BT(Instruction):
    """
    <instr>
        bt
    </instr>
    <syntax>
        bt reg, reg
        bt reg, const
    </syntax>
    <descr>
        Selects the bit in a bit string (specified with the first operand,
        called the bit base) at the bit-position designated by the bit offset
        (specified by the second operand) and stores the value of the bit in
        the CF flag. The bit base operand is a register; the bit offset operand
        can be a register or an immediate value. The instruction takes the
        modulo 16, 32, or 64 of the bit offset operand (modulo size depends on
        the mode and register size).
    </descr>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        # print('length', ops[1].get_val(line_num).bit_length())
        # if ops[1].get_val(line_num).bit_length() < 8:
        #     raise InvalidOperand('Operand 2 should be of size 8 bit')

        binary_num = bin(ops[0].get_val(line_num))
        index = len(binary_num) - ops[1].get_val(line_num)
        vmachine.flags["CF"] = binary_num[index]


class BTC(Instruction):
    """
    <instr>
        bt
    </instr>
    <syntax>
        bt reg, reg
        bt reg, const
    </syntax>
    <descr>
        Selects the bit in a bit string (specified with the first operand,
        called the bit base) at the bit-position desginated by the bit offset
        operand (second operand), stores the value of the bit in the CF flag,
        and complements the selected bit in the bit string. The bit base
        operand is a register; the bit offset operand can be a register or an
        immediate value. The instruction takes the modulo 16, 32, or 64 of the
        bit offset operand (modulo size depends on the mode and register size).
    </descr>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        # if ops[1].get_val(line_num).bit_length() != 8:
        #     raise InvalidOperand('Operand 2 should be of size 8 bit')

        binary_num = bin(ops[0].get_val(line_num))
        index = len(binary_num) - ops[1].get_val(line_num)
        if binary_num[index] == '1':
            complement = '0'
        else:
            complement = '1'
        vmachine.flags["CF"] = complement
